/* Copyright (c) 2005 David Hammerton - All rights reserved
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "tunesbrowser.h"

typedef struct
{
    /* tips */
    gboolean showtips;
    gint prevtip;
    gboolean tipshownonce;

    /* donate */
    gboolean showdonate;
} preferences;

static preferences prefs;

static void save_preferences();

static void set_defaults(preferences *pref)
{
    pref->showtips = TRUE;
    pref->prevtip = -1;
    pref->tipshownonce = FALSE;

    pref->showdonate = TRUE;
}

static void load_keys(preferences *pref, GKeyFile *keyfile)
{
    GError *error = NULL;
    gboolean boolres;
    gint intres;

    set_defaults(pref);

    boolres = g_key_file_get_boolean(keyfile, "tips", "showtips", &error);
    if (error == NULL) pref->showtips = boolres;
    intres = g_key_file_get_integer(keyfile, "tips", "prevtip", &error);
    if (error == NULL) pref->prevtip = intres;
    boolres = g_key_file_get_boolean(keyfile, "tips", "tipshownonce", &error);
    if (error == NULL) pref->tipshownonce = boolres;

    boolres = g_key_file_get_boolean(keyfile, "donate", "showdonate", &error);
    if (error == NULL) pref->showdonate = boolres;
}

static void save_keys(preferences *pref, GKeyFile *keyfile)
{
    g_key_file_set_boolean(keyfile, "tips", "showtips", pref->showtips);
    g_key_file_set_integer(keyfile, "tips", "prevtip", pref->prevtip);
    g_key_file_set_boolean(keyfile, "tips", "tipshownonce", pref->tipshownonce);

    g_key_file_set_boolean(keyfile, "donate", "showdonate", pref->showdonate);
}

gboolean pref_get_showtips() {
    return prefs.showtips;
}

void pref_set_showtip(gboolean v) {
    prefs.showtips = v;
    save_preferences();
}

gint pref_get_prevtip() {
    return prefs.prevtip;
}

void pref_set_prevtip(gint v) {
    prefs.prevtip = v;
    save_preferences();
}

gboolean pref_get_tipshownonce() {
    return prefs.tipshownonce;
}

void pref_set_tipshownonce(gboolean v) {
    prefs.tipshownonce = v;
    save_preferences();
}

gboolean pref_get_showdonate() {
    return prefs.showdonate;
}

void pref_set_showdonate(gboolean v) {
    prefs.showdonate = v;
    save_preferences();
}

static gchar *get_preference_filename()
{
    return g_build_filename(g_get_user_config_dir(),
                            "tunesbrowserrc",
                            NULL);
}

static void load_preferences()
{
    gboolean loaded;
    GKeyFile *keyfile;
    gchar *filename;

    set_defaults(&prefs);

    keyfile = g_key_file_new();

    filename = get_preference_filename();

    loaded = g_key_file_load_from_file(keyfile,
                                       filename,
                                       0,
                                       NULL);

    g_free(filename);

    if (!loaded) goto end;

    load_keys(&prefs, keyfile);


end:
    g_key_file_free(keyfile);
}

static void save_preferences()
{
    GKeyFile *keyfile;
    gchar *outfile = NULL;
    gsize outlen = 0;
    gchar *filename;
    GIOChannel *f;

    keyfile = g_key_file_new();

    save_keys(&prefs, keyfile);

    outfile = g_key_file_to_data(keyfile, &outlen, NULL);

    g_key_file_free(keyfile);

    filename = get_preference_filename();
    f = g_io_channel_new_file(filename, "w", NULL);
    g_free(filename);

    if (!f) goto end;

    g_io_channel_write_chars(f, outfile, outlen, NULL, NULL);

    g_io_channel_shutdown(f, TRUE, NULL);

end:
    g_free(outfile);
}

void init_preferences()
{
    load_preferences();
}

