/* Copyright (c) 2004 David Hammerton - All rights reserved
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "debug.h"
#include "lists.h"
#include "tunesbrowser.h"

#define DEFAULT_DEBUG_CHANNEL "sourcelist"

static void on_source_selection_changed(GtkTreeSelection *selection,
                                        gpointer data);

static GtkTreeStore *sourcelist;
static GtkTreeView *sourcetreeview;

static GdkPixbuf *remote_host_connected_image;
static GdkPixbuf *remote_host_notconnected_image;
static GdkPixbuf *playlist_image;

enum
{
    IMAGE_COLUMN,
    SOURCE_COLUMN,
    HOST_COLUMN,
    PLAYLIST_COLUMN,
    SOURCE_N_COLUMNS
};

#define XSTR(s) STR(s)
#define STR(s) #s
static void load_images()
{
    remote_host_connected_image = gdk_pixbuf_new_from_file(XSTR(UIDIR) "/comp.blue.png", NULL);
    remote_host_notconnected_image = gdk_pixbuf_new_from_file(XSTR(UIDIR) "/comp.red.png", NULL);
    playlist_image = gdk_pixbuf_new_from_file(XSTR(UIDIR) "/soundfile.png", NULL);
}
#undef XSTR
#undef STR

int sourcelist_init(GladeXML *xml)
{
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    GtkTreeSelection *select;

    load_images();

    sourcelist = gtk_tree_store_new (SOURCE_N_COLUMNS,
                                     GDK_TYPE_PIXBUF,
                                     G_TYPE_STRING,
                                     G_TYPE_POINTER,
                                     G_TYPE_POINTER);
    sourcetreeview = GTK_TREE_VIEW(glade_xml_get_widget(xml, "source"));

    gtk_tree_view_set_model(sourcetreeview, GTK_TREE_MODEL(sourcelist));

    renderer = gtk_cell_renderer_pixbuf_new();
    column = gtk_tree_view_column_new_with_attributes("",
                                                      renderer,
                                                      "pixbuf", IMAGE_COLUMN,
                                                      NULL);
    gtk_tree_view_append_column (sourcetreeview, column);

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes("Source",
                                                      renderer,
                                                      "text", SOURCE_COLUMN,
                                                      NULL);
    gtk_tree_view_append_column (sourcetreeview, column);

    select = gtk_tree_view_get_selection (GTK_TREE_VIEW (sourcetreeview));
    gtk_tree_selection_set_mode (select, GTK_SELECTION_SINGLE);
    g_signal_connect (G_OBJECT (select), "changed",
                      G_CALLBACK (on_source_selection_changed),
                      NULL);

    return 1;
}

static void sourcelist_append(daap_host *host, GtkTreeIter *iter_out)
{
    GtkTreeIter iter;
    GdkPixbuf *image;

    TRACE("(%p): %s\n", host, daap_host_get_sharename(host));

    gtk_tree_store_append(sourcelist, &iter, NULL);

    image = daap_host_is_connected(host) ?
        remote_host_connected_image : remote_host_notconnected_image;

    gtk_tree_store_set(sourcelist, &iter,
                       IMAGE_COLUMN, image,
                       SOURCE_COLUMN, daap_host_get_sharename(host),
                       HOST_COLUMN, host,
                       -1);

    if (iter_out) *iter_out = iter;
}

static void sourcelist_append_playlist(GtkTreeIter *parent_iter,
                                       daap_host *host,
                                       playlist *playlist,
                                       GtkTreeIter *iter_out)
{
    GtkTreeIter iter;

    TRACE("(%p:%p): %s\n", host, playlist,
          daap_host_get_playlist_name(host, playlist));

    gtk_tree_store_append(sourcelist, &iter, parent_iter);

    gtk_tree_store_set(sourcelist, &iter,
                       IMAGE_COLUMN, playlist_image,
                       SOURCE_COLUMN, daap_host_get_playlist_name(host, playlist),
                       HOST_COLUMN, host,
                       PLAYLIST_COLUMN, playlist,
                       -1);

    if (iter_out) *iter_out = iter;
}

typedef struct expandedrowsTAG
{
    daap_host *host;
    struct expandedrowsTAG *next;
} expandedrows;

static void expandedrows_cb(__UNUSED__ GtkTreeView *tv, GtkTreePath *path, gpointer pv_erows)
{
    GtkTreeIter iter;
    expandedrows **erows = (expandedrows**)pv_erows;

    if (gtk_tree_model_get_iter(GTK_TREE_MODEL(sourcelist), &iter, path))
    {
        void *iter_host = NULL;
        expandedrows *newrow = malloc(sizeof(expandedrows));

        gtk_tree_model_get(GTK_TREE_MODEL(sourcelist), &iter,
                           HOST_COLUMN, &iter_host,
                           -1);
        newrow->host = (daap_host*)iter_host;
        newrow->next = *erows;
        *erows = newrow;
    }
}

static int expanded_host(expandedrows *erows, daap_host *host)
{
    while (erows)
    {
        if (erows->host == host) return 1;
        erows = erows->next;
    }
    return 0;
}

static void expanded_free(expandedrows *erows)
{
    while (erows)
    {
        expandedrows *next = erows->next;
        free(erows);
        erows = next;
    }
}

void sourcelist_redraw()
{
    daap_host *cur = NULL;

    expandedrows *erows = NULL;
    daap_host *selected_host = NULL;
    playlist *selected_playlist = NULL;

    GtkTreeModel *model;
    GtkTreeIter selected_iter;
    GtkTreeSelection *selection = gtk_tree_view_get_selection(sourcetreeview);

    GtkTreePath *selected_path = NULL;

    TRACE("()\n");

    /* get expanded rows */
    gtk_tree_view_map_expanded_rows(sourcetreeview, expandedrows_cb, &erows);

    /* get selected row */
    if (gtk_tree_selection_get_selected(selection, &model, &selected_iter))
    {
        gtk_tree_model_get (model, &selected_iter,
                            HOST_COLUMN, &selected_host,
                            PLAYLIST_COLUMN, &selected_playlist,
                            -1);
    }

    gtk_tree_store_clear(sourcelist);

    while ((cur = daap_host_enumerate(cur)))
    {
        playlist *cur_playlist = NULL;
        GtkTreeIter iter;
        GtkTreePath *expanded_path = NULL;

        sourcelist_append(cur, &iter);

        /* if this is the previously selected one, reselect */
        if (cur == selected_host && selected_playlist == NULL)
        {
            selected_path = gtk_tree_model_get_path(GTK_TREE_MODEL(sourcelist),
                                                    &iter);
        }

        /* if this one was expanded, expand it */
        if (expanded_host(erows, cur))
        {
            expanded_path = gtk_tree_model_get_path(GTK_TREE_MODEL(sourcelist),
                                                    &iter);
        }

        while ((cur_playlist = daap_host_enum_playlists(cur, cur_playlist)))
        {
            GtkTreeIter iter_playlist;

            sourcelist_append_playlist(&iter, cur, cur_playlist, &iter_playlist);

            if (cur == selected_host && cur_playlist == selected_playlist)
            {
                selected_path = gtk_tree_model_get_path(GTK_TREE_MODEL(sourcelist),
                                                        &iter_playlist);
            }
        }

        if (expanded_path)
        {
            gtk_tree_view_expand_row(sourcetreeview, expanded_path, FALSE);
            gtk_tree_path_free(expanded_path);
        }
    }

    if (selected_path)
    {
        /* convert to iter */
        GtkTreeIter iter;
        if (gtk_tree_model_get_iter(GTK_TREE_MODEL(sourcelist), &iter, selected_path))
        {
            gtk_tree_selection_select_iter(selection, &iter);
        }
    }

    expanded_free(erows);
}

void sourcelist_set_source_connect_state(daap_host *host, int connected)
{
    GtkTreeIter iter;
    if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(sourcelist),
                                      &iter))
    {
        do
        {
            void *iter_host = NULL;
            gtk_tree_model_get(GTK_TREE_MODEL(sourcelist), &iter,
                               HOST_COLUMN, &iter_host,
                               -1);
            if (iter_host == host)
            {
                gtk_tree_store_set(sourcelist, &iter,
                                   IMAGE_COLUMN,
                                   connected ? remote_host_connected_image :
                                               remote_host_notconnected_image,
                                   -1);
            }
        } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(sourcelist),
                                          &iter));
    }
}

void sourcelist_set_sensitive(int sensitive)
{
    gtk_widget_set_sensitive(GTK_WIDGET(sourcetreeview),
                             sensitive ? TRUE : FALSE);
}

/* callback */
static void on_source_selection_changed(GtkTreeSelection *selection,
                                        __UNUSED__ gpointer data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;
    daap_host *host;
    playlist *playlist;


    if (gtk_tree_selection_get_selected(selection, &model, &iter))
    {
        gtk_tree_model_get (model, &iter,
                            HOST_COLUMN, &host,
                            PLAYLIST_COLUMN, &playlist,
                            -1);

        daap_host_set_visible(host, playlist);
    }

}



