/* Copyright (c) 2005 David Hammerton - All rights reserved
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "tunesbrowser.h"

#define DEFAULT_DEBUG_CHANNEL "tips"

typedef struct tipTAG
{
    char *tip_string;
    int n;
    struct tipTAG *prev;
    struct tipTAG *next;
} tip;

static tip *tips_head = NULL;
static tip *tips_tail = NULL;

static tip *visible_tip = NULL;

static GtkWindow *tipoftheday_win;
static GtkLabel *thetip;
static GtkCheckButton *tip_showonstart;

static void tips_read();

static tip *find_tip_by_id(int *id)
{
    tip *cur = NULL;
    int i;
    int n = 0;

    if (!id) return NULL;

    if (!tips_head) return NULL;

    cur = tips_head;
    for (i = 0; i < *id; i++)
    {
        if (cur->next)
        {
            cur = cur->next;
            n++;
        }
        else
        {
            cur = tips_head;
            n = 0;
            break;
        }
    }
    *id = n;
    return cur;
}

/* dialog box stuff */

static void set_tip(tip *settip)
{
    TRACE("tip %p (%s)\n", settip, settip->tip_string);
    gtk_label_set_markup(thetip, (gchar*)settip->tip_string);

    visible_tip = settip;

    pref_set_prevtip(settip->n);
}

static void on_tip_show_toggled(GtkToggleButton *togglebutton,
                                __UNUSED__ gpointer user_data)
{
    pref_set_showtip(gtk_toggle_button_get_active(togglebutton));
}

static void on_tip_prev(__UNUSED__ GtkButton *button, __UNUSED__ gpointer user_data)
{
    tip *prev = visible_tip->prev;
    if (!prev) prev = tips_tail;
    if (!prev) return;
    set_tip(prev);
}

static void on_tip_next(__UNUSED__ GtkButton *button, __UNUSED__ gpointer user_data)
{
    tip *next = visible_tip->next;
    if (!next) next = tips_head;
    if (!next) return;
    set_tip(next);
}

static void on_tip_close(__UNUSED__ GtkButton *button, __UNUSED__ gpointer user_data)
{
    gtk_widget_hide(GTK_WIDGET(tipoftheday_win));
}

void tips_show()
{
    int prev_tip = pref_get_prevtip();
    int next_tip;
    tip *cur;

    pref_set_tipshownonce(TRUE);

    if (!tips_head)
    {
        ERR("no tips loaded, can't show tips\n");
        return;
    }

    next_tip = prev_tip + 1;

    cur = find_tip_by_id(&next_tip);

    set_tip(cur);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tip_showonstart), pref_get_showtips());

    gtk_widget_show_all(GTK_WIDGET(tipoftheday_win));
}

void tips_init(GladeXML *xml)
{
    GtkButton *tip_next;
    GtkButton *tip_prev;
    GtkButton *tip_close;

    tipoftheday_win = GTK_WINDOW(glade_xml_get_widget(xml, "tipoftheday_win"));

    thetip = GTK_LABEL(glade_xml_get_widget(xml, "thetip"));

    tip_showonstart = GTK_CHECK_BUTTON(glade_xml_get_widget(xml, "tip_showonstart"));

    g_signal_connect(G_OBJECT(tip_showonstart), "toggled",
                     G_CALLBACK(on_tip_show_toggled), NULL);

    tip_prev = GTK_BUTTON(glade_xml_get_widget(xml, "tip_prev"));
    tip_next = GTK_BUTTON(glade_xml_get_widget(xml, "tip_next"));
    tip_close = GTK_BUTTON(glade_xml_get_widget(xml, "tip_close"));

    g_signal_connect(G_OBJECT(tip_prev), "clicked",
                     G_CALLBACK(on_tip_prev), NULL);
    g_signal_connect(G_OBJECT(tip_next), "clicked",
                     G_CALLBACK(on_tip_next), NULL);
    g_signal_connect(G_OBJECT(tip_close), "clicked",
                     G_CALLBACK(on_tip_close), NULL);

    /* make the close button have the focus */
    gtk_widget_grab_focus(GTK_WIDGET(tip_close));

    tips_read();
}

/* tips from file */

#define XSTR(s) STR(s)
#define STR(s) #s
static void tips_read()
{
    FILE *f;

    f = fopen(XSTR(UIDIR) "/tips.txt", "r");

    if (!f)
    {
        TRACE("could not load tip file\n");
        return;
    }

    while (!feof(f))
    {
        char buffer[1025] = {0}; /* max tip length */
        int len;

        tip *newtip;

        fgets(buffer, 1025, f);

        len = strlen(buffer);

        /* remove the \n (or \r\n or \r) if it exists */
        if (buffer[len-1] == '\n')
            buffer[--len] = '\0';
        if (buffer[len-1] == '\r')
            buffer[--len] = '\0';

        /* if tip is too big, abort */
        if (len == 1024 && !feof(f))
        {
            ERR("found a tip of > 1024 chars, aborting\n");
            break;
        }

        /* if it's a comment */
        if (buffer[0] == '#') continue;

        /* or if it's an empty line */
        if (strlen(buffer) == 0) continue;

        /* create new tip, add it to end of doubly linked list */
        newtip = malloc(sizeof(tip));
        newtip->tip_string = strdup(buffer);
        newtip->next = NULL;
        newtip->prev = tips_tail;
        if (tips_tail)
        {
            tips_tail->next = newtip;
            newtip->n = tips_tail->n + 1;
        }
        else
        {
            tips_head = newtip;
            newtip->n = 0;
        }
        tips_tail = newtip;
    }

    fclose(f);
}
#undef XSTR
#undef STR


